import dataclasses
import typing

from . import _indentlib
from . import string as _string


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tag:
    """
    Tag.

    ..  container:: example

        >>> abjad.Tag("YELLOW")
        Tag(string='YELLOW')

        >>> abjad.Tag("YELLOW:RED")
        Tag(string='YELLOW:RED')

        Raises exception on duplicate words in tag:

        >>> abjad.Tag("YELLOW:RED:RED")
        Traceback (most recent call last):
        ...
        Exception: duplicate words in tag: 'YELLOW:RED:RED'

        Raises exception on multiple only-edition tags:

        >>> abjad.Tag("+SEGMENT:+PARTS")
        Traceback (most recent call last):
            ...
        Exception: at most one only-edition tag: ['+SEGMENT', '+PARTS'].

        Raises exception on mixed only-edition / not-edition tags:

        >>> abjad.Tag("+SEGMENT:-PARTS")
        Traceback (most recent call last):
        ...
        Exception: only-edition and not-edition forbidden in same tag:
        <BLANKLINE>
          ['+SEGMENT'] / ['-PARTS']

    """

    string: str = ""

    def __post_init__(self):
        assert isinstance(self.string, str), repr(self.string)
        self.words()

    def append(self, word: "Tag") -> "Tag":
        """
        Appends ``word`` to tag.

        ..  container:: example

            >>> abjad.Tag("-PARTS").append(abjad.Tag("DEFAULT_CLEF"))
            Tag(string='-PARTS:DEFAULT_CLEF')

        """
        if not bool(word.string):
            return Tag(self.string)
        assert isinstance(word, Tag), repr(word)
        words = []
        if self.string:
            words.append(self.string)
        if word.string in self.words():
            raise Exception(f"{word} duplicates {self}.")
        words.append(word.string)
        string = ":".join(words)
        return Tag(string)

    def editions(self) -> list["Tag"]:
        """
        Gets edition tags in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").editions()
            []

            >>> abjad.Tag("+SEGMENT").only_edition()
            Tag(string='+SEGMENT')

            >>> abjad.Tag("+SEGMENT:FOO").only_edition()
            Tag(string='+SEGMENT')

            >>> abjad.Tag("-SEGMENT").editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:FOO").editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:-PARTS").editions()
            [Tag(string='-SEGMENT'), Tag(string='-PARTS')]

        """
        result = []
        for word in self.words():
            if word.startswith("+") or word.startswith("-"):
                result.append(Tag(word))
        return result

    def invert_edition_tags(self) -> "Tag":
        """
        Inverts edition tags in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").invert_edition_tags()
            Tag(string='FOO')

            >>> abjad.Tag("FOO:-PARTS").invert_edition_tags()
            Tag(string='FOO:+PARTS')

            >>> abjad.Tag("FOO:+PARTS").invert_edition_tags()
            Tag(string='FOO:-PARTS')

        """
        words = []
        for word in self.words():
            if word.startswith("+"):
                word_ = "-" + word[1:]
            elif word.startswith("-"):
                word_ = "+" + word[1:]
            else:
                word_ = word
            words.append(word_)
        string = ":".join(words)
        tag = Tag(string)
        return tag

    def not_editions(self) -> list["Tag"]:
        """
        Gets not-edition tags in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").not_editions()
            []

            >>> abjad.Tag("-SEGMENT").not_editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:FOO").not_editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:-PARTS").not_editions()
            [Tag(string='-SEGMENT'), Tag(string='-PARTS')]

        """
        result = []
        for word in self.words():
            if word.startswith("-"):
                result.append(Tag(word))
        return result

    def only_edition(self) -> typing.Optional["Tag"]:
        """
        Gets only-edition tag in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").only_edition() is None
            True

            >>> abjad.Tag("+SEGMENT").only_edition()
            Tag(string='+SEGMENT')

            >>> abjad.Tag("+SEGMENT:FOO").only_edition()
            Tag(string='+SEGMENT')

        """
        for word in self.words():
            if word.startswith("+"):
                return Tag(word)
        else:
            return None

    def retain_shoutcase(self) -> "Tag":
        """
        Retains shoutcase.

        ..  container:: example

            >>> tag = abjad.Tag("-PARTS:DEFAULT_CLEF:_apply_clef()")
            >>> tag.retain_shoutcase()
            Tag(string='-PARTS:DEFAULT_CLEF')

            >>> tag = abjad.Tag("_debug_function()")
            >>> tag.retain_shoutcase()
            Tag(string='')

        """
        words = []
        for word in self.words():
            if _string.is_shout_case(word) or word[0] in ("-", "+"):
                words.append(word)
        string = ":".join(words)
        return type(self)(string)

    def words(self) -> list[str]:
        """
        Gets words.

        ..  container:: example

            >>> abjad.Tag("-PARTS:DEFAULT_CLEF").words()
            ['-PARTS', 'DEFAULT_CLEF']

        """
        assert not self.string.startswith(":"), repr(self.string)
        words = self.string.split(":")
        assert isinstance(words, list), repr(words)
        words_ = []
        for word in words:
            if word in words_:
                raise Exception(f"duplicate words in tag: {self.string!r}")
            words_.append(word)
        only_edition_tags, not_edition_tags = [], []
        for word_ in words_:
            if word_.startswith("+"):
                only_edition_tags.append(word_)
            if word_.startswith("-"):
                not_edition_tags.append(word_)
        if 1 < len(only_edition_tags):
            raise Exception(f"at most one only-edition tag: {only_edition_tags!r}.")
        if only_edition_tags and not_edition_tags:
            message = "only-edition and not-edition forbidden in same tag:\n\n"
            message += f"  {only_edition_tags} / {not_edition_tags}"
            raise Exception(message)
        assert all(isinstance(_, str) for _ in words_), repr(words_)
        return words_


def _match_line(line, tag, current_tags):
    assert all(isinstance(_, Tag) for _ in current_tags), repr(current_tags)
    if tag in current_tags:
        return True
    if callable(tag):
        return tag(current_tags)
    assert isinstance(tag, Tag), repr(tag)
    return False


def activate(text: str, tag: Tag | typing.Callable) -> tuple[str, int, int]:
    r"""
    Activates ``tag`` in ``text``.

    ..  container:: example

        Writes (deactivated) tag with ``"%@%"`` prefix into LilyPond input:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup(r"\markup { \with-color #red Allegro }")
        >>> abjad.attach(
        ...     markup,
        ...     staff[0],
        ...     deactivate=True,
        ...     tag=abjad.Tag("RED_MARKUP"),
        ... )

        >>> text = abjad.lilypond(staff, tags=True)
        >>> text = abjad.tag.left_shift_tags(text)
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            %@% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

        Activates tag:

        >>> text, count, skipped = abjad.activate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro } %@%
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Deactivates tag again:

        >>> text, count, skipped = abjad.deactivate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            %@% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Activates tag again:

        >>> text, count, skipped = abjad.activate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro } %@%
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

    Tags can toggle indefinitely.

    Returns (text, count, skipped) triple.

    Count gives number of activated tags.

    Skipped gives number of skipped tags.
    """
    assert isinstance(text, str), repr(text)
    assert isinstance(tag, Tag) or callable(tag), repr(tag)
    lines: list[str] = []
    count, skipped_count = 0, 0
    treated_last_line = False
    found_already_active_on_last_line = False
    text_lines = text.split("\n")
    text_lines = [_ + "\n" for _ in text_lines[:-1]] + text_lines[-1:]
    lines = []
    current_tags = []
    for line in text_lines:
        if line.lstrip().startswith("%! "):
            lines.append(line)
            current_tag = Tag(line.strip()[3:])
            current_tags.append(current_tag)
            continue
        if not _match_line(line, tag, current_tags):
            lines.append(line)
            treated_last_line = False
            found_already_active_on_last_line = False
            current_tags = []
            continue
        first_nonwhitespace_index = len(line) - len(line.lstrip())
        index = first_nonwhitespace_index
        if line[index : index + 4] in ("%%% ", "%@% "):
            if "%@% " in line:
                line = line.replace("%@% ", "")
                suffix = " %@%"
            else:
                # TODO: replace with "" instead of "    "?
                line = line.replace("%%%", "   ")
                suffix = None
            assert line.endswith("\n"), repr(line)
            if suffix:
                line = line.strip("\n") + suffix + "\n"
            if not treated_last_line:
                count += 1
            treated_last_line = True
            found_already_active_on_last_line = False
        else:
            if not found_already_active_on_last_line:
                skipped_count += 1
            found_already_active_on_last_line = True
            treated_last_line = False
        lines.append(line)
        current_tags = []
    text = "".join(lines)
    return text, count, skipped_count


def deactivate(
    text: str,
    tag: Tag | typing.Callable,
    prepend_empty_chord: bool = False,
) -> tuple[str, int, int]:
    r"""
    Deactivates ``tag`` in ``text``.

    ..  container:: example

        Writes (active) tag into LilyPond input:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> string = r"\markup { \with-color #red Allegro }"
        >>> markup = abjad.Markup(string)
        >>> abjad.attach(
        ...     markup,
        ...     staff[0],
        ...     tag=abjad.Tag("RED_MARKUP"),
        ... )

        >>> text = abjad.lilypond(staff, tags=True)
        >>> text = abjad.tag.left_shift_tags(text)
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

        Deactivates tag:

        >>> text = abjad.lilypond(staff, tags=True)
        >>> text, count, skipped = abjad.deactivate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
        %%% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Activates tag again:

        >>> text, count, skipped = abjad.activate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Deactivates tag again:

        >>> text, count, skipped = abjad.deactivate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
        %%% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

    Tags can toggle indefinitely.
    """
    assert isinstance(text, str), repr(text)
    assert isinstance(tag, Tag) or callable(tag), repr(tag)
    count, skipped_count = 0, 0
    treated_last_line = False
    found_already_deactivated_on_last_line = False
    previous_line_was_tweak = False
    text_lines = text.split("\n")
    text_lines = [_ + "\n" for _ in text_lines[:-1]] + text_lines[-1:]
    lines, current_tags = [], []
    for line in text_lines:
        if line.lstrip().startswith("%! "):
            lines.append(line)
            current_tag = Tag(line.strip()[3:])
            current_tags.append(current_tag)
            continue
        if not _match_line(line, tag, current_tags):
            lines.append(line)
            treated_last_line = False
            found_already_deactivated_on_last_line = False
            current_tags = []
            continue
        start_column = len(line) - len(line.lstrip())
        if line[start_column] != "%":
            if " %@%" in line:
                prefix = "    " + "%@% "
                line = line.replace(" %@%", "")
            else:
                prefix = "%%% "
            if prepend_empty_chord and not previous_line_was_tweak:
                prefix += "<> "
            target = line[start_column - 4 : start_column]
            assert target == "    ", repr((line, target, start_column, tag))
            characters = list(line)
            characters[start_column - 4 : start_column] = list(prefix)
            line = "".join(characters)
            if not treated_last_line:
                count += 1
            treated_last_line = True
            found_already_deactivated_on_last_line = False
        else:
            if not found_already_deactivated_on_last_line:
                skipped_count += 1
            found_already_deactivated_on_last_line = True
            treated_last_line = False
        lines.append(line)
        previous_line_was_tweak = "tweak" in line
        current_tags = []
    text = "".join(lines)
    return text, count, skipped_count


def double_tag(strings: list[str], tag_: Tag, deactivate: bool = False) -> list[str]:
    """
    Double tags ``strings``.
    """
    assert all(isinstance(_, str) for _ in strings), repr(strings)
    assert isinstance(tag_, Tag), repr(tag_)
    half_indent = 2 * " "
    tag_lines = []
    if tag_.string:
        tag_lines_ = tag_.string.split(":")
        tag_lines_ = [half_indent + "%! " + _ for _ in tag_lines_]
        tag_lines.extend(tag_lines_)
        tag_lines.sort()
    if deactivate is True:
        strings = ["%@% " + _ for _ in strings]
    result = []
    for string in strings:
        if string.strip().startswith("%!"):
            result.append(string)
            continue
        result.extend(tag_lines)
        result.append(string)
    return result


def left_shift_tags(text: str) -> str:
    """
    Left shifts tags in ``strings``.
    """
    strings = text.split("\n")
    strings_ = []
    for string in strings:
        if "%@% " not in string or "%!" not in string:
            strings_.append(string)
            continue
        if not string.startswith(4 * " "):
            strings_.append(string)
            continue
        string_ = string[4:]
        tag_start = string_.find("%!")
        strings__ = list(string_)
        strings__[tag_start:tag_start] = _indentlib.INDENT
        string_ = "".join(strings__)
        strings_.append(string_)
    text = "\n".join(strings_)
    return text


def remove_tags(string: str) -> str:
    """
    Removes all tags from ``string``.
    """
    lines = []
    for line in string.split("\n"):
        if "%!" not in line:
            lines.append(line)
            continue
        tag_start = line.find("%!")
        line = line[:tag_start]
        line = line.rstrip()
        if line:
            lines.append(line)
    string = "\n".join(lines)
    return string



import copy

import pytest

import abjad


def test_Chord___copy___01():
    chord_1 = abjad.Chord("<ef' cs'' f''>4")
    chord_2 = copy.copy(chord_1)

    assert isinstance(chord_1, abjad.Chord)
    assert isinstance(chord_2, abjad.Chord)
    assert abjad.lilypond(chord_1) == abjad.lilypond(chord_2)
    assert chord_1 is not chord_2


def test_Chord___copy___02():
    """
    Chord copies LilyPond duration multiplier.
    """

    chord_1 = abjad.Chord("<ef' cs'' f''>4 * 1/2")
    chord_2 = copy.copy(chord_1)

    assert isinstance(chord_1, abjad.Chord)
    assert isinstance(chord_2, abjad.Chord)
    assert abjad.lilypond(chord_1) == abjad.lilypond(chord_2)
    assert chord_1 is not chord_2


def test_Chord___copy___03():
    """
    Chord copies LilyPond grob overrides and LilyPond context settings.
    """

    chord_1 = abjad.Chord("<ef' cs'' f''>4")
    abjad.override(chord_1).Staff.NoteHead.color = "#red"
    abjad.override(chord_1).Accidental.color = "#red"
    abjad.setting(chord_1).tupletFullLength = True
    chord_2 = copy.copy(chord_1)

    assert isinstance(chord_1, abjad.Chord)
    assert isinstance(chord_2, abjad.Chord)
    assert abjad.lilypond(chord_1) == abjad.lilypond(chord_2)
    assert chord_1 is not chord_2


def test_Chord___copy___04():
    """
    Chord copies tweaked note-heads.
    """

    chord_1 = abjad.Chord("<c' e' g'>4")
    abjad.tweak(chord_1.note_heads[0], r"\tweak color #red")
    chord_2 = copy.copy(chord_1)

    assert abjad.lilypond(chord_1) == abjad.string.normalize(
        r"""
        <
            \tweak color #red
            c'
            e'
            g'
        >4
        """
    )

    assert abjad.lilypond(chord_2) == abjad.string.normalize(
        r"""
        <
            \tweak color #red
            c'
            e'
            g'
        >4
        """
    )

    assert abjad.lilypond(chord_1) == abjad.lilypond(chord_2)
    assert chord_1 is not chord_2

    assert chord_1.note_heads[0] == chord_2.note_heads[0]
    assert chord_1.note_heads[1] == chord_2.note_heads[1]
    assert chord_1.note_heads[2] == chord_2.note_heads[2]

    assert chord_1.note_heads[0] is not chord_2.note_heads[0]
    assert chord_1.note_heads[1] is not chord_2.note_heads[1]
    assert chord_1.note_heads[2] is not chord_2.note_heads[2]


def test_Chord___copy___05():
    """
    Chord coipes articulations and markup.
    """

    chord_1 = abjad.Chord("<ef' cs'' f''>4")
    articulation_1 = abjad.Articulation("staccato")
    abjad.attach(articulation_1, chord_1)
    markup_1 = abjad.Markup(r"\markup foo")
    abjad.attach(markup_1, chord_1)

    chord_2 = copy.copy(chord_1)

    assert abjad.lilypond(chord_1) == abjad.lilypond(chord_2)
    assert chord_1 is not chord_2

    articulation_2 = abjad.get.indicators(chord_2, abjad.Articulation)[0]
    assert articulation_1 == articulation_2
    assert articulation_1 is not articulation_2

    markup_2 = abjad.get.markup(chord_2)[0]
    assert markup_1 == markup_2
    assert markup_1 is not markup_2


def test_Chord___deepcopy___01():
    """
    Chord deepchopies note-heads.
    """

    chord_1 = abjad.Chord("<c' e' g'>4")
    abjad.tweak(chord_1.note_heads[0], r"\tweak color #red")
    chord_2 = copy.deepcopy(chord_1)

    assert abjad.lilypond(chord_1) == abjad.string.normalize(
        r"""
        <
            \tweak color #red
            c'
            e'
            g'
        >4
        """
    )

    assert abjad.lilypond(chord_2) == abjad.string.normalize(
        r"""
        <
            \tweak color #red
            c'
            e'
            g'
        >4
        """
    )

    assert abjad.lilypond(chord_1) == abjad.lilypond(chord_2)
    assert chord_1 is not chord_2

    assert chord_1.note_heads[0] == chord_2.note_heads[0]
    assert chord_1.note_heads[1] == chord_2.note_heads[1]
    assert chord_1.note_heads[2] == chord_2.note_heads[2]

    assert chord_1.note_heads[0] is not chord_2.note_heads[0]
    assert chord_1.note_heads[1] is not chord_2.note_heads[1]
    assert chord_1.note_heads[2] is not chord_2.note_heads[2]


def test_Chord___eq____01():
    """
    True only when chords have the same object ID.
    """

    chord_1 = abjad.Chord([0, 4, 7], (1, 4))
    chord_2 = abjad.Chord([0, 4, 7], (1, 4))
    chord_3 = abjad.Chord([0, 4, 6], (1, 4))

    assert chord_1 == chord_1
    assert not chord_1 == chord_2
    assert not chord_1 == chord_3
    assert not chord_2 == chord_1
    assert chord_2 == chord_2
    assert not chord_2 == chord_3
    assert not chord_3 == chord_1
    assert not chord_3 == chord_2
    assert chord_3 == chord_3


def test_Chord___init___01():
    """
    Initialize empty chord.
    """

    chord = abjad.Chord([], (1, 4))
    assert abjad.lilypond(chord) == "<>4"


def test_Chord___init___02():
    """
    Initialize chord with pitch numbers.
    """

    chord = abjad.Chord([2, 4, 5], (1, 4))
    assert abjad.lilypond(chord) == "<d' e' f'>4"


def test_Chord___init___03():
    """
    Initialize chord with pitch tokens.
    """

    chord = abjad.Chord([("ds", 4), ("ef", 4)], (1, 4))
    assert abjad.lilypond(chord) == "<ds' ef'>4"


def test_Chord___init___04():
    """
    Initialize chord with pitches.
    """

    pitches = []
    pitches.append(abjad.NamedPitch("D#4"))
    pitches.append(abjad.NamedPitch("Eb4"))
    chord = abjad.Chord(pitches, (1, 4))
    assert abjad.lilypond(chord) == "<ds' ef'>4"


def test_Chord___init___05():
    """
    Initialize chord with pitches and pitch numbers together.
    """

    pitches = [2, ("ef", 4), abjad.NamedPitch(4)]
    chord = abjad.Chord(pitches, (1, 4))
    assert abjad.lilypond(chord) == "<d' ef' e'>4"


def test_Chord___init___06():
    """
    Initialize chord with list of pitch names.
    """

    pitches = ["d'", "ef'", "e'"]
    chord = abjad.Chord(pitches, (1, 4))
    assert abjad.lilypond(chord) == "<d' ef' e'>4"


def test_Chord___init___07():
    """
    Initialize chord with LilyPond input string.
    """

    chord = abjad.Chord("<d' ef' e'>4")
    assert abjad.lilypond(chord) == "<d' ef' e'>4"


def test_Chord___init___08():
    """
    Initialize chord from skip.
    """

    skip = abjad.Skip("s8")
    chord = abjad.Chord(skip)

    assert abjad.lilypond(skip) == "s8"
    assert abjad.lilypond(chord) == "<>8"

    assert abjad.wf.wellformed(skip)
    assert abjad.wf.wellformed(chord)


def test_Chord___init___09():
    """
    Initialize chord from tupletized skip.
    """

    tuplet = abjad.Tuplet((2, 3), "s8 s8 s8")
    chord = abjad.Chord(tuplet[0])

    assert abjad.lilypond(chord) == "<>8"
    assert abjad.get.parentage(chord).parent is None
    assert abjad.wf.wellformed(chord)


def test_Chord___init___10():
    """
    Initialize chord from containerized skip.
    """

    tuplet = abjad.Voice("s8 s8 s8")
    chord = abjad.Chord(tuplet[0])

    assert abjad.lilypond(chord) == "<>8"
    assert abjad.get.parentage(chord).parent is None
    assert abjad.wf.wellformed(chord)


def test_Chord___init___11():
    """
    Initialize chord from beamed skip.
    """

    staff = abjad.Staff("c'8 [ s8 c'8 ]")
    chord = abjad.Chord(staff[1])

    assert abjad.lilypond(chord) == "<>8"
    assert abjad.get.parentage(chord).parent is None
    assert abjad.wf.wellformed(chord)


def test_Chord___init___12():
    """
    Initialize chord from rest.
    """

    rest = abjad.Rest("r8")
    chord = abjad.Chord(rest)

    assert abjad.lilypond(rest) == "r8"
    assert abjad.lilypond(chord) == "<>8"
    assert abjad.wf.wellformed(rest)
    assert abjad.wf.wellformed(chord)


def test_Chord___init___13():
    """
    Initialize chord from tupletized rest.
    """

    tuplet = abjad.Tuplet((2, 3), "r8 r8 r8")
    chord = abjad.Chord(tuplet[1])

    assert abjad.lilypond(chord) == "<>8"
    assert abjad.wf.wellformed(chord)
    assert abjad.get.parentage(chord).parent is None


def test_Chord___init___14():
    """
    Initialize chord from note.
    """

    note = abjad.Note("d'8")
    chord = abjad.Chord(note)

    assert abjad.lilypond(note) == "d'8"
    assert abjad.lilypond(chord) == "<d'>8"
    assert abjad.wf.wellformed(note)
    assert abjad.wf.wellformed(chord)


def test_Chord___init___15():
    """
    Initialize chord from tupletized note.
    """

    tuplet = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    chord = abjad.Chord(tuplet[1])

    assert abjad.lilypond(chord) == "<c'>8"
    assert abjad.wf.wellformed(chord)
    assert abjad.get.parentage(chord).parent is None


def test_Chord___init___16():
    """
    Initialize chord from spanned note.
    """

    staff = abjad.Staff("c'8 ( d'8 e'8 f'8 )")
    chord = abjad.Chord(staff[1])

    assert abjad.lilypond(chord) == "<d'>8"
    assert abjad.wf.wellformed(chord)
    assert abjad.get.parentage(chord).parent is None


def test_Chord___init___17():
    """
    Initialize empty chord from LilyPond input string.
    """

    chord = abjad.Chord("<>8.")

    assert abjad.lilypond(chord) == "<>8."
    assert not len(chord.note_heads)


def test_Chord___init___18():
    """
    Initialize chord from LilyPond input string with forced and cautionary
    accidentals.
    """

    chord = abjad.Chord("<c!? e? g! b>4")

    assert abjad.lilypond(chord) == "<c!? e? g! b>4"


def test_Chord___init___19():
    """
    Initialize chord from note with forced and cautionary accidentals.
    """

    note = abjad.Note("c'!?4")
    chord = abjad.Chord(note)

    assert abjad.lilypond(chord) == "<c'!?>4"


def test_Chord___init___20():
    """
    Initialize chord from other chord.
    """

    chord_1 = abjad.Chord("<c' e' g' bf'>4")
    chord_2 = abjad.Chord(chord_1, abjad.Duration(1, 8))

    assert abjad.lilypond(chord_2) == "<c' e' g' bf'>8"


def test_Chord___init___21():
    """
    Initialize chord with drum pitches.
    """

    chord = abjad.Chord("<sn? bd! tamb>4")

    assert abjad.lilypond(chord) == "<bassdrum! snare? tambourine>4"


def test_Chord_written_pitches_01():
    """
    Returns immutable tuple of pitches in chord.
    """

    chord = abjad.Chord("<d' e' f'>4")
    pitches = chord.written_pitches

    assert isinstance(pitches, tuple)
    assert len(pitches) == 3
    with pytest.raises(Exception):
        pitches.pop()
    with pytest.raises(Exception):
        pitches.remove(pitches[0])


def test_Chord_written_pitches_02():
    """
    Equivalent written pitches compare equal.
    """

    chord_1 = abjad.Chord("<d' e' f'>4")
    chord_2 = abjad.Chord("<d' e' f'>4")

    assert chord_1.written_pitches == chord_2.written_pitches


def test_Chord_written_pitches_03():
    """
    Set written pitches with pitch numbers.
    """

    chord = abjad.Chord([], (1, 4))
    chord.written_pitches = [4, 3, 2]
    assert abjad.lilypond(chord) == "<d' ef' e'>4"

    chord.written_pitches = (4, 3, 2)
    assert abjad.lilypond(chord) == "<d' ef' e'>4"


def test_Chord_written_pitches_04():
    """
    Set written pitches with pitches.
    """

    chord = abjad.Chord([], (1, 4))
    chord.written_pitches = [
        abjad.NamedPitch(4),
        abjad.NamedPitch(3),
        abjad.NamedPitch(2),
    ]

    assert abjad.lilypond(chord) == "<d' ef' e'>4"


def test_Chord_written_pitches_05():
    """
    Set written pitches with both pitches and pitch numbers.
    """

    chord = abjad.Chord([], (1, 4))
    chord.written_pitches = [4, abjad.NamedPitch(3), abjad.NamedPitch(2)]

    assert abjad.lilypond(chord) == "<d' ef' e'>4"



import copy

import abjad


def test_Rest___cmp___01():
    rest_1 = abjad.Rest((1, 4))
    rest_2 = abjad.Rest((1, 4))
    rest_3 = abjad.Rest((1, 8))

    assert not rest_1 == rest_2
    assert not rest_1 == rest_3
    assert not rest_2 == rest_3


def test_Rest___cmp___02():
    rest_1 = abjad.Rest((1, 4))
    rest_2 = abjad.Rest((1, 4))
    rest_3 = abjad.Rest((1, 8))

    assert rest_1 != rest_2
    assert rest_1 != rest_3
    assert rest_2 != rest_3


def test_Rest___copy___01():
    """
    Copy rest.
    """

    rest_1 = abjad.Rest((1, 4))
    rest_2 = copy.copy(rest_1)

    assert isinstance(rest_1, abjad.Rest)
    assert isinstance(rest_2, abjad.Rest)
    assert abjad.lilypond(rest_1) == abjad.lilypond(rest_2)
    assert rest_1 is not rest_2


def test_Rest___copy___02():
    """
    Copy rest with LilyPond multiplier.
    """

    rest_1 = abjad.Rest("r4", multiplier=(1, 2))
    rest_2 = copy.copy(rest_1)

    assert isinstance(rest_1, abjad.Rest)
    assert isinstance(rest_2, abjad.Rest)
    assert abjad.lilypond(rest_1) == abjad.lilypond(rest_2)
    assert rest_1 is not rest_2


def test_Rest___copy___03():
    """
    Copy rest with LilyPond grob overrides and LilyPond context settings.
    """

    rest_1 = abjad.Rest((1, 4))
    abjad.override(rest_1).Staff.NoteHead.color = "#red"
    abjad.override(rest_1).Accidental.color = "#red"
    abjad.setting(rest_1).tupletFullLength = True
    rest_2 = copy.copy(rest_1)

    assert isinstance(rest_1, abjad.Rest)
    assert isinstance(rest_2, abjad.Rest)
    assert abjad.lilypond(rest_1) == abjad.lilypond(rest_2)
    assert rest_1 is not rest_2


def test_Rest___init___01():
    """
    Initialize rest from LilyPond input string.
    """

    rest = abjad.Rest("r8.")

    assert rest.written_duration == abjad.Duration(3, 16)


def test_Rest___init___02():
    """
    Initialize rest from other rest.
    """

    rest_1 = abjad.Rest("r4", multiplier=(1, 2))
    abjad.override(rest_1).Staff.NoteHead.color = "#red"
    rest_2 = abjad.Rest(rest_1)

    assert isinstance(rest_1, abjad.Rest)
    assert isinstance(rest_2, abjad.Rest)
    assert abjad.lilypond(rest_1) == abjad.lilypond(rest_2)
    assert rest_1 is not rest_2


def test_Rest___init___03():
    """
    Initialize rest from chord.
    """

    chord = abjad.Chord([2, 3, 4], (1, 4))
    rest = abjad.Rest(chord)

    assert isinstance(rest, abjad.Rest)
    assert dir(chord) == dir(abjad.Chord([2, 3, 4], (1, 4)))
    assert dir(rest) == dir(abjad.Rest((1, 4)))
    assert rest.written_duration == chord.written_duration


def test_Rest___init___04():
    """
    Initialize rest from tupletized chord.
    """

    chord = abjad.Chord([2, 3, 4], abjad.Duration(1, 4))
    chords = abjad.mutate.copy(chord, 3)
    tuplet = abjad.Tuplet((2, 3), chords)
    rest = abjad.Rest(tuplet[0])

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___05():
    """
    Initialize rest from beamed chord.
    """

    chord = abjad.Chord([2, 3, 4], abjad.Duration(1, 8))
    chords = abjad.mutate.copy(chord, 3)
    voice = abjad.Voice(chords)
    abjad.beam(voice[:])
    rest = abjad.Rest(voice[0])

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r8
        [
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___06():
    """
    Initialize rest from skip.
    """

    skip = abjad.Skip("s4")
    rest = abjad.Rest(skip)

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___07():
    """
    Initialize rest from tupletted skip.
    """

    tuplet = abjad.Tuplet((2, 3), "s4 s4 s4")
    rest = abjad.Rest(tuplet[0])

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___08():
    """
    Initialize rest from beamed skip.
    """

    staff = abjad.Staff("c'8 [ s4 c'd ]")
    rest = abjad.Rest(staff[1])

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___09():
    """
    Initialize rest from note.
    """

    note = abjad.Note("c'4")
    rest = abjad.Rest(note)

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___10():
    """
    Initialize rest from tupletized note.
    """

    tuplet = abjad.Tuplet((2, 3), "c'4 d'4 e'4")
    rest = abjad.Rest(tuplet[0])

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___11():
    """
    Initialize rest from beamed note.
    """

    staff = abjad.Staff("c'8 [ d'8 e'8 ]")
    rest = abjad.Rest(staff[0])

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r8
        [
        """
    )

    assert abjad.wf.wellformed(rest)


def test_Rest___init___12():
    """
    Initialize multiple rests from spanned notes.
    """

    voice = abjad.Voice("c'8 ( d'8 e'8 f'8 )")
    for note in voice:
        rest = abjad.Rest(note)
        abjad.mutate.replace(note, rest)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            r8
            (
            r8
            r8
            r8
            )
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Rest___init___13():
    """
    Initializes rest from empty input.
    """

    rest = abjad.Rest()

    assert abjad.lilypond(rest) == abjad.string.normalize(
        r"""
        r4
        """
    )

    assert abjad.wf.wellformed(rest)



import pytest

import abjad


def test_makers__group_by_implied_prolation_01():
    with pytest.raises(Exception):
        abjad.makers._group_by_implied_prolation([])


def test_makers__group_by_implied_prolation_02():
    fractions = [(1, 4)]
    duration = abjad.makers._group_by_implied_prolation(fractions)
    assert duration == [[(1, 4)]]


def test_makers__group_by_implied_prolation_03():
    fractions = [(1, 4), (1, 4), (1, 8)]
    duration = abjad.makers._group_by_implied_prolation(fractions)
    assert duration == [
        [
            (1, 4),
            (1, 4),
            (1, 8),
        ]
    ]


def test_makers__group_by_implied_prolation_04():
    fractions = [(1, 4), (1, 3), (1, 8)]
    duration = abjad.makers._group_by_implied_prolation(fractions)
    assert duration == [
        [(1, 4)],
        [(1, 3)],
        [(1, 8)],
    ]


def test_makers__group_by_implied_prolation_05():
    fractions = [(1, 4), (1, 2), (1, 3)]
    duration = abjad.makers._group_by_implied_prolation(fractions)
    assert duration == [
        [(1, 4), (1, 2)],
        [(1, 3)],
    ]


def test_makers__group_by_implied_prolation_06():
    fractions = [(1, 4), (1, 2), (1, 3), (1, 6), (1, 5)]
    duration = abjad.makers._group_by_implied_prolation(fractions)
    assert duration == [
        [(1, 4), (1, 2)],
        [(1, 3), (1, 6)],
        [(1, 5)],
    ]


def test_makers__group_by_implied_prolation_07():
    fractions = [(1, 24), (2, 24), (3, 24), (4, 24), (5, 24), (6, 24)]
    duration = abjad.makers._group_by_implied_prolation(fractions)
    assert duration == [
        [
            (1, 24),
            (2, 24),
            (3, 24),
            (4, 24),
            (5, 24),
            (6, 24),
        ]
    ]


def test_makers_tuplet_from_ratio_and_pair_01():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 4), (6, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 7/6
        {
            c'16
            c'8
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_02():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 1, 2, 4), (6, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 4/3
        {
            c'16
            c'16
            c'8
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_03():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((-2, 3, 7), (7, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 12/7
        {
            r8
            c'8.
            c'4..
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_04():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((7, 7, -4, -1), (1, 4))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tuplet 19/16
        {
            c'16..
            c'16..
            r16
            r64
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_05():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 2), (12, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'8
            c'4
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_06():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((2, 4, 4), (12, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'8
            c'4
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_07():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((4, 8, 8), (12, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/3
        {
            c'4
            c'2
            c'2
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_08():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((8, 16, 16), (12, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/3
        {
            c'4
            c'2
            c'2
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_09():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((2, 4, 4), (3, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/3
        {
            c'16
            c'8
            c'8
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_10():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((2, 4, 4), (6, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/3
        {
            c'8
            c'4
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_11():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((2, 4, 4), (12, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'8
            c'4
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_12():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((2, 4, 4), (24, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'4
            c'2
            c'2
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_13():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 2), (6, 2))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'2
            c'1
            c'1
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_14():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 2), (6, 4))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'4
            c'2
            c'2
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_15():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 2), (6, 8))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'8
            c'4
            c'4
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_16():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 2), (6, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 5/6
        {
            c'16
            c'8
            c'8
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_17():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, -1, -1), (3, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 1/1
        {
            c'16
            r16
            r16
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_18():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 1, -1, -1), (4, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 1/1
        {
            c'16
            c'16
            r16
            r16
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_19():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 1, 1, -1, -1), (5, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 1/1
        {
            c'16
            c'16
            c'16
            r16
            r16
        }
        """
    )


def test_makers_tuplet_from_ratio_and_pair_20():
    tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 1, 1, 1, -1, -1), (6, 16))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 1/1
        {
            c'16
            c'16
            c'16
            c'16
            r16
            r16
        }
        """
    )



"""
Updates start offsets, stop offsets and indicators everywhere in score.

..  note:: This is probably the most important part of Abjad to optimize. Use the
    profiler to figure out how many unnecessary updates are happening. Then reimplement.
    As a hint, _updatelib.py implements a weird version of the "observer pattern."
    It may make sense to revisit a textbook example of the observer pattern and review
    the implementation of _updatelib.py.

"""

import fractions

from . import duration as _duration
from . import indicators as _indicators
from . import iterate as _iterate
from . import math as _math
from . import obgc as _obgc
from . import parentage as _parentage
from . import score as _score
from . import sequence as _sequence
from . import timespan as _timespan


def _get_after_grace_leaf_offsets(leaf):
    container = leaf._parent
    main_leaf = container._main_leaf
    main_leaf_stop_offset = main_leaf._stop_offset
    assert main_leaf_stop_offset is not None
    displacement = -leaf._get_duration()
    sibling = leaf._sibling(1)
    while sibling is not None and sibling._parent is container:
        displacement -= sibling._get_duration()
        sibling = sibling._sibling(1)
    if leaf._parent is not None and leaf._parent._main_leaf is not None:
        main_leaf = leaf._parent._main_leaf
        sibling = main_leaf._sibling(1)
        if (
            sibling is not None
            and hasattr(sibling, "_before_grace_container")
            and sibling._before_grace_container is not None
        ):
            before_grace_container = sibling._before_grace_container
            duration = before_grace_container._get_duration()
            displacement -= duration
    start_offset = _duration.Offset(main_leaf_stop_offset, displacement=displacement)
    displacement += leaf._get_duration()
    stop_offset = _duration.Offset(main_leaf_stop_offset, displacement=displacement)
    return start_offset, stop_offset


def _get_before_grace_leaf_offsets(leaf):
    container = leaf._parent
    main_leaf = container._main_leaf
    main_leaf_start_offset = main_leaf._start_offset
    assert main_leaf_start_offset is not None
    displacement = -leaf._get_duration()
    sibling = leaf._sibling(1)
    while sibling is not None and sibling._parent is container:
        displacement -= sibling._get_duration()
        sibling = sibling._sibling(1)
    start_offset = _duration.Offset(main_leaf_start_offset, displacement=displacement)
    displacement += leaf._get_duration()
    stop_offset = _duration.Offset(main_leaf_start_offset, displacement=displacement)
    return start_offset, stop_offset


def _get_independent_after_grace_leaf_offsets(leaf):
    container = leaf._parent
    main_leaf = container._sibling(-1)
    main_leaf_stop_offset = main_leaf._stop_offset
    assert main_leaf_stop_offset is not None
    displacement = -leaf._get_duration()
    sibling = leaf._sibling(1)
    while sibling is not None and sibling._parent is container:
        displacement -= sibling._get_duration()
        sibling = sibling._sibling(1)
    """
    if leaf._parent is not None and leaf._parent._sibling(-1) is not None:
        main_leaf = leaf._parent._sibling(-1)
        sibling = main_leaf._sibling(1)
        if (
            sibling is not None
            and hasattr(sibling, "_before_grace_container")
            and sibling._before_grace_container is not None
        ):
            before_grace_container = sibling._before_grace_container
            duration = before_grace_container._get_duration()
            displacement -= duration
    """
    start_offset = _duration.Offset(main_leaf_stop_offset, displacement=displacement)
    displacement += leaf._get_duration()
    stop_offset = _duration.Offset(main_leaf_stop_offset, displacement=displacement)
    return start_offset, stop_offset


def _get_measure_start_offsets(component):
    wrappers = []
    prototype = _indicators.TimeSignature
    root = _parentage.Parentage(component).root
    for component_ in _iterate_entire_score(root):
        wrappers_ = component_._get_indicators(prototype, unwrap=False)
        wrappers.extend(wrappers_)
    pairs = []
    for wrapper in wrappers:
        component = wrapper.component
        start_offset = component._get_timespan().start_offset
        time_signature = wrapper.unbundle_indicator()
        pair = start_offset, time_signature
        pairs.append(pair)
    offset_zero = _duration.Offset(0)
    default_time_signature = _indicators.TimeSignature((4, 4))
    default_pair = (offset_zero, default_time_signature)
    if pairs and not pairs[0] == offset_zero:
        pairs.insert(0, default_pair)
    elif not pairs:
        pairs = [default_pair]
    pairs.sort(key=lambda x: x[0])
    score_stop_offset = root._get_timespan().stop_offset
    dummy_last_pair = (score_stop_offset, None)
    pairs.append(dummy_last_pair)
    measure_start_offsets = []
    at_first_measure = True
    for current_pair, next_pair in _sequence.nwise(pairs):
        current_start_offset, current_time_signature = current_pair
        next_start_offset, next_time_signature = next_pair
        measure_start_offset = current_start_offset
        while measure_start_offset < next_start_offset:
            measure_start_offsets.append(measure_start_offset)
            partial = current_time_signature.partial
            if at_first_measure and partial is not None:
                measure_start_offset += partial
                measure_start_offsets.append(measure_start_offset)
                at_first_measure = False
            measure_start_offset += current_time_signature.duration
    return measure_start_offsets


def _get_obgc_leaf_offsets(leaf):
    obgc = leaf._parent
    assert type(obgc).__name__ == "OnBeatGraceContainer", repr(obgc)
    first_nongrace_leaf = obgc.get_first_nongrace_leaf()
    first_nongrace_leaf_start_offset = first_nongrace_leaf._start_offset
    assert first_nongrace_leaf_start_offset is not None
    first_nongrace_leaf_start_offset = _duration.Offset(
        first_nongrace_leaf_start_offset.pair
    )
    start_displacement = _duration.Duration(0)
    sibling = leaf._sibling(-1)
    while sibling is not None and sibling._parent is obgc:
        start_displacement += sibling._get_duration()
        sibling = sibling._sibling(-1)
    stop_displacement = start_displacement + leaf._get_duration()
    if start_displacement == 0:
        start_displacement = None
    start_offset = _duration.Offset(
        first_nongrace_leaf_start_offset.pair, displacement=start_displacement
    )
    stop_offset = _duration.Offset(
        first_nongrace_leaf_start_offset.pair, displacement=stop_displacement
    )
    return start_offset, stop_offset


def _get_score_tree_state_flags(parentage):
    offsets_are_current = True
    indicators_are_current = True
    offsets_in_seconds_are_current = True
    for component in parentage:
        if offsets_are_current:
            if not component._offsets_are_current:
                offsets_are_current = False
        if indicators_are_current:
            if not component._indicators_are_current:
                indicators_are_current = False
        if offsets_in_seconds_are_current:
            if not component._offsets_in_seconds_are_current:
                offsets_in_seconds_are_current = False
    return (
        offsets_are_current,
        indicators_are_current,
        offsets_in_seconds_are_current,
    )


def _iterate_entire_score(root):
    """
    NOTE: RETURNS GRACE NOTES LAST (AND OUT-OF-ORDER).
    """
    components = list(_iterate.components(root, grace=False))
    graces = _iterate.components(root, grace=True)
    components.extend(graces)
    return components


def _make_metronome_mark_map(root):
    pairs = []
    all_stop_offsets = set()
    for component in _iterate_entire_score(root):
        indicators = component._get_indicators(_indicators.MetronomeMark)
        if len(indicators) == 1:
            metronome_mark = indicators[0]
            if not metronome_mark.is_imprecise:
                pair = (component._start_offset, metronome_mark)
                pairs.append(pair)
        if component._stop_offset is not None:
            all_stop_offsets.add(component._stop_offset)
    pairs.sort(key=lambda _: _[0])
    if not pairs:
        return
    if pairs[0][0] != 0:
        return
    score_stop_offset = max(all_stop_offsets)
    timespans = _timespan.TimespanList()
    clocktime_start_offset = _duration.Offset(0)
    for left, right in _sequence.nwise(pairs, wrapped=True):
        metronome_mark = left[-1]
        start_offset = left[0]
        stop_offset = right[0]
        # last timespan
        if stop_offset == 0:
            stop_offset = score_stop_offset
        duration = stop_offset - start_offset
        multiplier = fractions.Fraction(60, metronome_mark.units_per_minute)
        clocktime_duration = duration / metronome_mark.reference_duration
        clocktime_duration *= multiplier
        timespan = _timespan.Timespan(
            start_offset=start_offset,
            stop_offset=stop_offset,
            annotation=(clocktime_start_offset, clocktime_duration),
        )
        timespans.append(timespan)
        clocktime_start_offset += clocktime_duration
    return timespans


# TODO: reimplement with some type of bisection
def _to_measure_number(component, measure_start_offsets):
    component_start_offset = component._get_timespan().start_offset
    displacement = component_start_offset.displacement
    if displacement is not None:
        component_start_offset = _duration.Offset(
            component_start_offset, displacement=None
        )
        # score-initial grace music only:
        if displacement < 0 and component_start_offset == 0:
            measure_number = 0
            return measure_number
    measure_start_offsets = measure_start_offsets[:]
    measure_start_offsets.append(_math.Infinity())
    pairs = _sequence.nwise(measure_start_offsets)
    for measure_index, pair in enumerate(pairs):
        if pair[0] <= component_start_offset < pair[-1]:
            measure_number = measure_index + 1
            return measure_number
    message = f"can not find measure number for {repr(component)}:\n"
    message += f"   {repr(measure_start_offsets)}"
    raise ValueError(message)


def _update_all_indicators(root):
    """
    Updating indicators does not update offsets.
    On the other hand, getting an effective indicator does update
    offsets when at least one indicator of the appropriate type
    attaches to score.
    """
    components = _iterate_entire_score(root)
    for component in components:
        for wrapper in component._get_indicators(unwrap=False):
            if wrapper.context is not None:
                wrapper._update_effective_context()
        component._indicators_are_current = True


def _update_all_offsets(root):
    """
    Updating offsets does not update indicators.
    Updating offsets does not update offsets in seconds.
    """
    obgc_components = []
    for component in _iterate_entire_score(root):
        if isinstance(component, _obgc.OnBeatGraceContainer) or isinstance(
            component._parent, _obgc.OnBeatGraceContainer
        ):
            obgc_components.append(component)
        else:
            _update_component_offsets(component)
            component._offsets_are_current = True
    for component in obgc_components:
        _update_component_offsets(component)
        component._offsets_are_current = True


def _update_all_offsets_in_seconds(root):
    _update_all_offsets(root)
    timespans = _make_metronome_mark_map(root)
    for component in _iterate_entire_score(root):
        _update_clocktime_offsets(component, timespans)
        component._offsets_in_seconds_are_current = True


def _update_clocktime_offsets(component, timespans):
    if not timespans:
        return
    for timespan in timespans:
        if timespan.start_offset <= component._start_offset < timespan.stop_offset:
            pair = timespan.annotation
            clocktime_start_offset, clocktime_duration = pair
            local_offset = component._start_offset - timespan.start_offset
            multiplier = local_offset / timespan.duration
            duration = multiplier * clocktime_duration
            offset = clocktime_start_offset + duration
            component._start_offset_in_seconds = _duration.Offset(offset)
        if timespan.start_offset <= component._stop_offset < timespan.stop_offset:
            pair = timespan.annotation
            clocktime_start_offset, clocktime_duration = pair
            local_offset = component._stop_offset - timespan.start_offset
            multiplier = local_offset / timespan.duration
            duration = multiplier * clocktime_duration
            offset = clocktime_start_offset + duration
            component._stop_offset_in_seconds = _duration.Offset(offset)
            return
    if component._stop_offset == timespans[-1].stop_offset:
        pair = timespans[-1].annotation
        clocktime_start_offset, clocktime_duration = pair
        offset = clocktime_start_offset + clocktime_duration
        component._stop_offset_in_seconds = _duration.Offset(offset)
        return
    raise Exception(f"can not find {offset} in {timespans}.")


def _update_component_offsets(component):
    if isinstance(component, _score.BeforeGraceContainer):
        pair = _get_before_grace_leaf_offsets(component[0])
        start_offset = pair[0]
        pair = _get_before_grace_leaf_offsets(component[-1])
        stop_offset = pair[-1]
    elif isinstance(component._parent, _score.BeforeGraceContainer):
        pair = _get_before_grace_leaf_offsets(component)
        start_offset, stop_offset = pair
    elif isinstance(component, _obgc.OnBeatGraceContainer):
        pair = _get_obgc_leaf_offsets(component[0])
        start_offset = pair[0]
        pair = _get_obgc_leaf_offsets(component[-1])
        stop_offset = pair[-1]
    elif isinstance(component._parent, _obgc.OnBeatGraceContainer):
        pair = _get_obgc_leaf_offsets(component)
        start_offset, stop_offset = pair
    elif isinstance(component, _score.AfterGraceContainer):
        pair = _get_after_grace_leaf_offsets(component[0])
        start_offset = pair[0]
        pair = _get_after_grace_leaf_offsets(component[-1])
        stop_offset = pair[-1]
    elif isinstance(component, _score.IndependentAfterGraceContainer):
        pair = _get_independent_after_grace_leaf_offsets(component[0])
        start_offset = pair[0]
        pair = _get_independent_after_grace_leaf_offsets(component[-1])
        stop_offset = pair[-1]
    elif isinstance(component._parent, _score.AfterGraceContainer):
        pair = _get_after_grace_leaf_offsets(component)
        start_offset, stop_offset = pair
    elif isinstance(component._parent, _score.IndependentAfterGraceContainer):
        pair = _get_independent_after_grace_leaf_offsets(component)
        start_offset, stop_offset = pair
    else:
        previous = component._sibling(-1)
        if previous is not None:
            start_offset = previous._stop_offset
        else:
            start_offset = _duration.Offset(0)
        # on-beat anchor leaf:
        if (
            component._parent is not None
            and _obgc._is_obgc_nongrace_voice(component._parent)
            and component is component._parent[0]
        ):
            nongrace_voice = component._parent
            assert _obgc._is_obgc_nongrace_voice(nongrace_voice)
            obgc = None
            polyphony_container = nongrace_voice._parent
            assert _obgc._is_obgc_polyphony_container(polyphony_container)
            index = polyphony_container.index(nongrace_voice)
            # TODO: limit OBGC to index 0 (not index 1) in polyphony container
            if index == 0:
                obgc = polyphony_container[1]
            else:
                obgc = polyphony_container[0]
            if obgc is not None:
                grace_leaf_durations = [_._get_duration() for _ in obgc]
                start_displacement = sum(grace_leaf_durations)
                start_offset = _duration.Offset(
                    start_offset, displacement=start_displacement
                )
        stop_offset = start_offset + component._get_duration()
    component._start_offset = start_offset
    component._stop_offset = stop_offset
    component._timespan.start_offset = start_offset
    component._timespan.stop_offset = stop_offset


def _update_measure_numbers(component):
    measure_start_offsets = _get_measure_start_offsets(component)
    root = _parentage.Parentage(component).root
    for component in _iterate_entire_score(root):
        measure_number = _to_measure_number(component, measure_start_offsets)
        component._measure_number = measure_number


def _update_now(component, offsets=False, offsets_in_seconds=False, indicators=False):
    assert offsets or offsets_in_seconds or indicators
    parentage = component._get_parentage()
    for component_ in parentage:
        if component_._is_forbidden_to_update:
            return
    (
        offsets_are_current,
        indicators_are_current,
        offsets_in_seconds_are_current,
    ) = _get_score_tree_state_flags(parentage)
    root = parentage[-1]
    if offsets and not offsets_are_current:
        _update_all_offsets(root)
    if offsets_in_seconds and not offsets_in_seconds_are_current:
        _update_all_offsets_in_seconds(root)
    if indicators and not indicators_are_current:
        _update_all_indicators(root)



"""
Enumerations.
"""

import enum


class Comparison(enum.Enum):
    """
    Enumeration of amount comparisons.
    """

    LESS = -1
    EXACT = 0
    MORE = 1


class Horizontal(enum.Enum):
    """
    Enumeration of horizontal alignments.
    """

    LEFT = -1
    MIDDLE = 0
    RIGHT = 1


class Vertical(enum.Enum):
    """
    Enumeration of vertical alignments.
    """

    DOWN = -1
    CENTER = 0
    UP = 1


CENTER = Vertical.CENTER
DOWN = Vertical.DOWN
EXACT = Comparison.EXACT
LEFT = Horizontal.LEFT
LESS = Comparison.LESS
MIDDLE = Horizontal.MIDDLE
MORE = Comparison.MORE
RIGHT = Horizontal.RIGHT
UP = Vertical.UP



import abjad


def test_ly_environment_01():
    assert abjad.lyenv.contexts is not None
    print(abjad.lyenv.contexts)

    assert abjad.lyenv.current_module is not None
    print(abjad.lyenv.current_module)

    assert abjad.lyenv.engravers is not None
    print(abjad.lyenv.engravers)

    assert abjad.lyenv.grob_interfaces is not None
    print(abjad.lyenv.grob_interfaces)

    assert abjad.lyenv.interface_properties is not None
    print(abjad.lyenv.interface_properties)

    assert abjad.lyenv.language_pitch_names is not None
    print(abjad.lyenv.language_pitch_names)

    assert abjad.lyenv.markup_functions is not None
    print(abjad.lyenv.markup_functions)

    assert abjad.lyenv.markup_list_functions is not None
    print(abjad.lyenv.markup_list_functions)

    assert abjad.lyenv.music_glyphs is not None
    print(abjad.lyenv.music_glyphs)
